#!/usr/bin/env python3
# SPDX-License-Identifier: LGPL-2.1-or-later

"""
Check out pkg/{distribution}.
With -u, fetch commits, and if changed, commit the latest hash.
"""

import argparse
import json
import shlex
import subprocess
from pathlib import Path

def parse_args():
    p = argparse.ArgumentParser(
        description=__doc__,
    )
    p.add_argument(
        'distribution',
        nargs='+',
    )
    p.add_argument(
        '--no-fetch',
        dest='fetch',
        action='store_false',
        default=True,
    )
    p.add_argument(
        '--update', '-u',
        action='store_true',
        default=False,
    )
    return p.parse_args()

def read_config(distro: str):
    cmd = ['mkosi', '--json', '-d', distro, 'summary']
    print(f"+ {shlex.join(cmd)}")
    text = subprocess.check_output(cmd, text=True)

    data = json.loads(text)
    images = {image["Image"]: image for image in data["Images"]}
    return images["build"]

def commit_file(distro: str, file: Path, commit: str, changes: str):
    message = '\n'.join((
        f'mkosi: update {distro} commit reference',
        '',
        changes))

    cmd = ['git', 'commit', '-m', message, str(file)]
    print(f"+ {shlex.join(cmd)}")
    subprocess.check_call(cmd)

def checkout_distro(args, distro: str, config: dict):
    dest = Path(f'pkg/{distro}')
    if dest.exists():
        print(f'{dest} already exists.')
        return

    url = config['Environment']['GIT_URL']
    branch = config['Environment']['GIT_BRANCH']

    # Only debian uses source-git for now…
    reference = [f'--reference-if-able=.'] if distro == 'debian' else []

    cmd = [
        'git', 'clone', url,
        f'--branch={branch}',
        dest.as_posix(),
        *reference,
    ]
    print(f"+ {shlex.join(cmd)}")
    subprocess.check_call(cmd)

    args.fetch = False  # no need to fetch if we just cloned

def update_distro(args, distro: str, config: dict):
    branch = config['Environment']['GIT_BRANCH']
    old_commit = config['Environment']['GIT_COMMIT']

    cmd = ['git', '-C', f'pkg/{distro}', 'switch', branch]
    print(f"+ {shlex.join(cmd)}")
    subprocess.check_call(cmd)

    cmd = ['git', '-C', f'pkg/{distro}', 'fetch', 'origin', '-v',
           f'{branch}:remotes/origin/{branch}']
    print(f"+ {shlex.join(cmd)}")
    subprocess.check_call(cmd)

    cmd = ['git', '-C', f'pkg/{distro}', 'rev-parse', f'refs/remotes/origin/{branch}']
    print(f"+ {shlex.join(cmd)}")
    new_commit = subprocess.check_output(cmd, text=True).strip()

    if old_commit == new_commit:
        print(f'{distro}: commit {new_commit!s} is still fresh')
        return

    cmd = ['git', '-C', f'pkg/{distro}', 'log', '--graph', '--first-parent',
           '--pretty=oneline', '--no-decorate', '--abbrev-commit', '--abbrev=10',
           f'{old_commit}..{new_commit}']
    print(f"+ {shlex.join(cmd)}")
    changes = subprocess.check_output(cmd, text=True).strip()

    conf_dir = Path('mkosi.images/build/mkosi.conf.d')
    files = conf_dir.glob('*/*.conf')
    for file in files:
        s = file.read_text()
        if old_commit in s:
            print(f'{distro}: {file}: found old hash, updating…')
            new = s.replace(old_commit, new_commit)
            assert new != s
            file.write_text(new)
            commit_file(distro, file, new_commit, changes)
            break
    else:
        raise ValueError(f'{distro}: hash {new_commit} not found under {conf_dir}')

if __name__ == '__main__':
    args = parse_args()

    for distro in args.distribution:
        config = read_config(distro)
        checkout_distro(args, distro, config)
        if args.update:
            update_distro(args, distro, config)



#!/usr/bin/env python3
# SPDX-License-Identifier: LGPL-2.1-or-later

import json
import re
import subprocess
import sys

def main():
    build_dir = sys.argv[1]

    out = subprocess.run(["meson", "introspect", "--installed", build_dir],
                         stdout=subprocess.PIPE, check=True)
    files = json.loads(out.stdout)
    for file in sorted(files.values()):
        if re.search("^/usr/lib/systemd/(system|user)/", file) and not file.endswith(".conf"):
            print(file)

if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# SPDX-License-Identifier: LGPL-2.1-or-later

"""
Check out mkosi into specified location.
With -u, if changed, commit the latest hash.
"""

import argparse
import shlex
import subprocess
import re
from pathlib import Path

URL = 'https://github.com/systemd/mkosi'
BRANCH = 'main'  # We only want to ever use commits on upstream 'main' branch
FILENAME = Path('.github/workflows/mkosi.yml')

def parse_args():
    p = argparse.ArgumentParser(
        description=__doc__,
    )
    p.add_argument(
        'dir',
        type=Path,
    )
    p.add_argument(
        '--update', '-u',
        action='store_true',
        default=False,
    )
    return p.parse_args()

def read_config():
    print(f'Reading {FILENAME}…')
    matches = [m.group(1)
               for line in open(FILENAME)
               if (m := re.match('^- uses: systemd/mkosi@([a-z0-9]{40})$',
                                 line.strip()))]
    assert len(matches) == 1
    return matches[0]

def commit_file(args, file: Path, commit: str, changes: str):
    cmd = [
        'git', '-C', args.dir.as_posix(),
        'describe',
        '--always',
        commit]
    print(f"+ {shlex.join(cmd)}")
    desc = subprocess.check_output(cmd, text=True).strip()

    message = '\n'.join((
        f'mkosi: update mkosi commit reference to {desc}',
        '',
        changes))

    cmd = ['git', 'commit', '-m', message, file.as_posix()]
    print(f"+ {shlex.join(cmd)}")
    subprocess.check_call(cmd)

def checkout_mkosi(args):
    if args.dir.exists():
        print(f'{args.dir} already exists.')
        return

    cmd = [
        'git', 'clone', URL,
        f'--branch={BRANCH}',
        args.dir.as_posix(),
    ]
    print(f"+ {shlex.join(cmd)}")
    subprocess.check_call(cmd)

def update_mkosi(args):
    old_commit = read_config()

    cmd = ['git', '-C', args.dir.as_posix(), 'rev-parse', f'refs/remotes/origin/{BRANCH}']
    print(f"+ {shlex.join(cmd)}")
    new_commit = subprocess.check_output(cmd, text=True).strip()

    if old_commit == new_commit:
        print(f'mkosi: commit {new_commit!s} is still fresh')
        return

    cmd = ['git', '-C', args.dir.as_posix(), 'log', '--graph',
           '--pretty=oneline', '--no-decorate', '--abbrev-commit', '--abbrev=10',
           f'{old_commit}..{new_commit}']
    print(f"+ {shlex.join(cmd)}")
    changes = subprocess.check_output(cmd, text=True).strip()

    s = FILENAME.read_text()
    assert old_commit in s
    print(f'mkosi: {FILENAME}: found old hash, updating…')
    new = s.replace(old_commit, new_commit)
    assert new != s
    FILENAME.write_text(new)
    commit_file(args, FILENAME, new_commit, changes)

if __name__ == '__main__':
    args = parse_args()
    checkout_mkosi(args)
    if args.update:
        update_mkosi(args)



#!/usr/bin/env python3
# SPDX-License-Identifier: MIT-0
#
# Implement the systemd notify protocol without external dependencies.
# Supports both readiness notification on startup and on reloading,
# according to the protocol defined at:
# https://www.freedesktop.org/software/systemd/man/latest/sd_notify.html
# This protocol is guaranteed to be stable as per:
# https://systemd.io/PORTABILITY_AND_STABILITY/

import errno
import os
import signal
import socket
import sys
import time

reloading = False
terminating = False

def notify(message):
    if not message:
        raise ValueError("notify() requires a message")

    socket_path = os.environ.get("NOTIFY_SOCKET")
    if not socket_path:
        return

    if socket_path[0] not in ("/", "@"):
        raise OSError(errno.EAFNOSUPPORT, "Unsupported socket type")

    # Handle abstract socket.
    if socket_path[0] == "@":
        socket_path = "\0" + socket_path[1:]

    with socket.socket(socket.AF_UNIX, socket.SOCK_DGRAM | socket.SOCK_CLOEXEC) as sock:
        sock.connect(socket_path)
        sock.sendall(message)

def notify_ready():
    notify(b"READY=1")

def notify_reloading():
    microsecs = time.clock_gettime_ns(time.CLOCK_MONOTONIC) // 1000
    notify(f"RELOADING=1\nMONOTONIC_USEC={microsecs}".encode())

def notify_stopping():
    notify(b"STOPPING=1")

def reload(signum, frame):
    global reloading
    reloading = True

def terminate(signum, frame):
    global terminating
    terminating = True

def main():
    print("Doing initial setup")
    global reloading, terminating

    # Set up signal handlers.
    print("Setting up signal handlers")
    signal.signal(signal.SIGHUP, reload)
    signal.signal(signal.SIGINT, terminate)
    signal.signal(signal.SIGTERM, terminate)

    # Do any other setup work here.

    # Once all setup is done, signal readiness.
    print("Done setting up")
    notify_ready()

    print("Starting loop")
    while not terminating:
        if reloading:
            print("Reloading")
            reloading = False

            # Support notifying the manager when reloading configuration.
            # This allows accurate state tracking as well as automatically
            # enabling 'systemctl reload' without needing to manually
            # specify an ExecReload= line in the unit file.

            notify_reloading()

            # Do some reconfiguration work here.

            print("Done reloading")
            notify_ready()

        # Do the real work here ...

        print("Sleeping for five seconds")
        time.sleep(5)

    print("Terminating")
    notify_stopping()

if __name__ == "__main__":
    sys.stdout.reconfigure(line_buffering=True)
    print("Starting app")
    main()
    print("Stopped app")



#!/usr/bin/env python3
# SPDX-License-Identifier: LGPL-2.1-or-later
#
# ruff: noqa: E501 UP015

import os
import re
import sys
from typing import IO


def process_sym_file(file: IO[str]) -> None:
    for line in file:
        m = re.search(r'^ +([a-zA-Z0-9_]+);', line)
        if m:
            if m[1] == 'sd_bus_object_vtable_format':
                print(f'        {{ "{m[1]}", &{m[1]} }},')
            else:
                print(f'        {{ "{m[1]}", {m[1]} }},')


def process_header_file(file: IO[str]) -> None:
    for line in file:
        if (
            line.startswith('#')
            or line.startswith('typedef')
            or line.startswith('extern "C"')
            or line.startswith('__extension__')
            or line.startswith('/*')
            or ' __inline__ ' in line
            or re.search(r'^\s+', line)
        ):
            continue

        m = re.search(r'^(.*)\s*__attribute__', line)
        if m:
            line = m[1]

        m = re.search(r'^(.*)\s*_sd_printf_', line)
        if m:
            line = m[1]

        # Functions
        m = re.search(r'^(\S+\s+)+\**(\w+)\s*\(', line)
        if m:
            print(f'        {{ "{m[2]}", {m[2]} }},')
            continue

        # Variables
        m = re.search(r'^extern\s', line)
        if m:
            n = line.split()[-1].rstrip(';')
            print(f'        {{ "{n}", &{n} }},')
            continue

        # Functions defined by macro
        m = re.search(r'_SD_DEFINE_POINTER_CLEANUP_FUNC\(\w+,\s*(\w+)\)', line)
        if m:
            print(f'        {{ "{m[1]}", {m[1]} }},')
            continue


def process_source_file(file: IO[str]) -> None:
    for line in file:
        # Functions
        m = re.search(r'^_public_\s+(\S+\s+)+\**(\w+)\s*\(', line)
        if m:
            print(f'        {{ "{m[2]}", {m[2]} }},')
            continue

        # Variables
        m = re.search(r'^_public_\s+(\S+\s+)+\**(\w+)\s*=', line)
        if m:
            print(f'        {{ "{m[2]}", &{m[2]} }},')
            continue

        # Functions defined through a macro
        m = re.search(r'^DEFINE_PUBLIC_TRIVIAL_REF_FUNC\([^,]+,\s*(\w+)\s*\)', line)
        if m:
            print(f'        {{ "{m[1]}_ref", {m[1]}_ref }},')
            continue

        m = re.search(r'^DEFINE_PUBLIC_TRIVIAL_UNREF_FUNC\([^,]+,\s*(\w+)\s*,', line)
        if m:
            print(f'        {{ "{m[1]}_unref", {m[1]}_unref }},')
            continue

        m = re.search(r'^DEFINE_PUBLIC_TRIVIAL_REF_UNREF_FUNC\([^,]+,\s*(\w+)\s*,', line)
        if m:
            print(f'        {{ "{m[1]}_ref", {m[1]}_ref }},')
            print(f'        {{ "{m[1]}_unref", {m[1]}_unref }},')
            continue

        m = re.search(r'^_DEFINE_STRING_TABLE_LOOKUP\((\w+),\s*\w+,\s*_public_\s*\)', line)
        if m:
            print(f'        {{ "{m[1]}_from_string", {m[1]}_from_string }},')
            print(f'        {{ "{m[1]}_to_string", {m[1]}_to_string }},')
            continue


print("""/* SPDX-License-Identifier: LGPL-2.1-or-later */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

""")

for header in sys.argv[3:]:
    print('#include "{}"'.format(header.split('/')[-1]))

print("""
/* We want to check deprecated symbols too, without complaining */
#pragma GCC diagnostic ignored "-Wdeprecated-declarations"
""")

print("""
struct symbol {
        const char *name;
        const void *symbol;
};
static struct symbol symbols_from_sym[] = {""")

with open(sys.argv[1], 'r') as f:
    process_sym_file(f)

print("""        {}
}, symbols_from_header[] = {""")

for header in sys.argv[3:]:
    with open(header, 'r') as f:
        process_header_file(f)

print("""        {}
}, symbols_from_source[] = {""")

for dirpath, _, filenames in sorted(os.walk(sys.argv[2])):
    for filename in sorted(filenames):
        if not filename.endswith('.c') and not filename.endswith('.h'):
            continue
        with open(os.path.join(dirpath, filename), 'r') as f:
            process_source_file(f)

print("""        {}
};

static int sort_callback(const void *a, const void *b) {
        const struct symbol *x = a, *y = b;
        return strcmp(x->name, y->name);
}

int main(void) {
        size_t size = sizeof(symbols_from_sym[0]),
                n_sym = sizeof(symbols_from_sym)/sizeof(symbols_from_sym[0]) - 1,
                n_header = sizeof(symbols_from_header)/sizeof(symbols_from_header[0]) - 1,
                n_source = sizeof(symbols_from_source)/sizeof(symbols_from_source[0]) - 1;

        qsort(symbols_from_sym, n_sym, size, sort_callback);
        qsort(symbols_from_header, n_header, size, sort_callback);
        qsort(symbols_from_source, n_source, size, sort_callback);

        puts("From symbol file:");
        for (size_t i = 0; i < n_sym; i++)
                printf("%p: %s\\n", symbols_from_sym[i].symbol, symbols_from_sym[i].name);

        puts("\\nFrom header files:");
        for (size_t i = 0; i < n_header; i++)
                printf("%p: %s\\n", symbols_from_header[i].symbol, symbols_from_header[i].name);

        puts("\\nFrom source files:");
        for (size_t i = 0; i < n_source; i++)
                printf("%p: %s\\n", symbols_from_source[i].symbol, symbols_from_source[i].name);

        puts("");
        printf("Found %zu symbols from symbol file.\\n", n_sym);
        printf("Found %zu symbols from header files.\\n", n_header);
        printf("Found %zu symbols from source files.\\n", n_source);

        unsigned n_error = 0;

        for (size_t i = 0; i < n_sym; i++) {
                if (!bsearch(symbols_from_sym+i, symbols_from_header, n_header, size, sort_callback)) {
                        printf("Found in symbol file, but not in headers: %s\\n", symbols_from_sym[i].name);
                        n_error++;
                }
                if (!bsearch(symbols_from_sym+i, symbols_from_source, n_source, size, sort_callback)) {
                        printf("Found in symbol file, but not in sources: %s\\n", symbols_from_sym[i].name);
                        n_error++;
                }
        }

        for (size_t i = 0; i < n_header; i++) {
                if (!bsearch(symbols_from_header+i, symbols_from_sym, n_sym, size, sort_callback)) {
                        printf("Found in header file, but not in symbol file: %s\\n", symbols_from_header[i].name);
                        n_error++;
                }
                if (!bsearch(symbols_from_header+i, symbols_from_source, n_source, size, sort_callback)) {
                        printf("Found in header file, but not in sources: %s\\n", symbols_from_header[i].name);
                        n_error++;
                }
        }

        for (size_t i = 0; i < n_source; i++) {
                if (!bsearch(symbols_from_source+i, symbols_from_sym, n_sym, size, sort_callback)) {
                        printf("Found in source file, but not in symbol file: %s\\n", symbols_from_source[i].name);
                        n_error++;
                }
                if (!bsearch(symbols_from_source+i, symbols_from_header, n_header, size, sort_callback)) {
                        printf("Found in source file, but not in header: %s\\n", symbols_from_source[i].name);
                        n_error++;
                }
        }

        return n_error == 0 ? EXIT_SUCCESS : EXIT_FAILURE;
}""")
